package ssw.mj.codegen;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;

/* MicroJava Instruction Decoder
   =============================
*/
public class Decoder {

	private static byte[] code;	// code buffer
	private static int cur;			// address of next byte to decode
	private static int adr;			// address of currently decoded instruction

	private static int get() { return code[cur++]; }
	private static int get2() { return (get()<<8) + (get() & 0xFF); }
	private static int get4() { return (get2()<<16) + (get2() & 0xFFFF); }

	private static String jumpDist() {
		int dist = get2();
		int pos = adr + dist;
		return String.valueOf(dist) + " (=" + String.valueOf(pos) + ")";
	}

	public static void decode(byte[] buf, int off, int len) {
		code = buf;
		cur = off;
		adr = cur;
		while (cur < len) {
			System.out.println(adr + ": " + decode(get()));
			adr = cur;
		}
	}

	public static String decode (int opCode) {
		String s = "";
		switch (opCode) {
			case 1: s = "load " + get(); break;
			case 2: s = "load_0"; break;
			case 3: s = "load_1"; break;
			case 4: s = "load_2"; break;
			case 5: s = "load_3"; break;
			case 6: s = "store " + get(); break;
			case 7: s = "store_0"; break;
			case 8: s = "store_1"; break;
			case 9: s = "store_2"; break;
			case 10: s = "store_3"; break;
			case 11: s = "getstatic " + get2(); break;
			case 12: s = "putstatic " + get2(); break;
			case 13: s = "getfield " + get2(); break;
			case 14: s = "putfield " + get2(); break;
			case 15: s = "const_0"; break;
			case 16: s = "const_1"; break;
			case 17: s = "const_2"; break;
			case 18: s = "const_3"; break;
			case 19: s = "const_4"; break;
			case 20: s = "const_5"; break;
			case 21: s = "const_m1"; break;
			case 22: s = "const " + get4(); break;
			case 23: s = "add"; break;
			case 24: s = "sub"; break;
			case 25: s = "mul"; break;
			case 26: s = "div"; break;
			case 27: s = "rem"; break;
			case 28: s = "neg"; break;
			case 29: s = "shl"; break;
			case 30: s = "shr"; break;
			case 31: s = "inc " + get() + "," + get(); break;
			case 32: s = "new " + get2(); break;
			case 33: s = "newarray " + get(); break;
			case 34: s = "aload"; break;
			case 35: s = "astore"; break;
			case 36: s = "baload"; break;
			case 37: s = "bastore"; break;
			case 38: s = "arraylength"; break;
			case 39: s = "pop"; break;
			case 40: s = "dup"; break;
			case 41: s = "dup2"; break;
			case 42: s = "jmp " + jumpDist(); break;
			case 43: s = "jeq " + jumpDist(); break;
			case 44: s = "jne " + jumpDist(); break;
			case 45: s = "jlt " + jumpDist(); break;
			case 46: s = "jle " + jumpDist(); break;
			case 47: s = "jgt " + jumpDist(); break;
			case 48: s = "jge " + jumpDist(); break;
			case 49: s = "call " + jumpDist(); break;
			case 50: s = "return"; break;
			case 51: s = "enter " + get() + "," + get(); break;
			case 52: s = "exit"; break;
			case 53: s = "read"; break;
			case 54: s = "print"; break;
			case 55: s = "bread"; break;
			case 56: s = "bprint"; break;
			case 57: s = "trap " + get(); break;
			default: s = "-- error--"; break;
		}
		return s;
	}

	public static void decodeFile (String filename) {
		try {
			DataInputStream in = new DataInputStream(
					new BufferedInputStream(new FileInputStream(filename)));
			byte[] sig = new byte[2];
			in.read(sig, 0, 2);
			System.out.println("" + (char)sig[0] + (char)sig[1]);
			int codeSize = in.readInt();
			System.out.println("codesize = " + codeSize);
			System.out.println("datasize = " + in.readInt());
			System.out.println("startPC  = " + in.readInt());
			byte[] code = new byte[codeSize];
			in.read(code);
			decode(code, 0, codeSize);
		} catch (IOException ioe) {
			System.out.println("IOException");
		}
	}

	public static void main (String[] args) {
		if (args.length > 0) decodeFile(args[0]);
	}
}
