package ssw.mj.test;

import junit.framework.TestCase;

import ssw.mj.codegen.Code;
import ssw.mj.codegen.Item;
import ssw.mj.symtab.Tab;

public class CodeTest extends TestCase {
	TestByteArrayOutputStream code;

	protected void setUp () { 
		Code.init(); 
		code = new TestByteArrayOutputStream();
	}
	
	private void checkInt2ByteArray (int x, byte[] e) {
		byte[] a = TestByteArrayOutputStream.int2ByteArray(x);
		assertEquals(x + ": byte 0", e[0], a[0]);
		assertEquals(x + ": byte 1", e[1], a[1]);
		assertEquals(x + ": byte 2", e[2], a[2]);
		assertEquals(x + ": byte 3", e[3], a[3]);
	}
	
	public void testInt2ByteArray () {
		byte[] expect;

		expect = new byte[] {(byte)0xff, (byte)0xff, (byte)0xff, (byte)0xff };
		checkInt2ByteArray(-1, expect);
		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00 };
		checkInt2ByteArray(0, expect);			// 2^0
		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x01 };
		checkInt2ByteArray(1, expect);

		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x7f };
		checkInt2ByteArray(127, expect);
		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x80 };
		checkInt2ByteArray(128, expect);		// 2^7
		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x81 };
		checkInt2ByteArray(129, expect);

		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0x00, (byte)0xff };
		checkInt2ByteArray(255, expect);
		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0x01, (byte)0x00 };
		checkInt2ByteArray(256, expect);		// 2^8
		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0x01, (byte)0x01 };
		checkInt2ByteArray(257, expect);

		expect = new byte[] {(byte)0x00, (byte)0x00, (byte)0xff, (byte)0xff };
		checkInt2ByteArray(65535, expect);
		expect = new byte[] {(byte)0x00, (byte)0x01, (byte)0x00, (byte)0x00 };
		checkInt2ByteArray(65536, expect);		// 2^16
		expect = new byte[] {(byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01 };
		checkInt2ByteArray(65537, expect);

		expect = new byte[] {(byte)0x00, (byte)0xff, (byte)0xff, (byte)0xff };
		checkInt2ByteArray(16777215, expect);
		expect = new byte[] {(byte)0x01, (byte)0x00, (byte)0x00, (byte)0x00 };
		checkInt2ByteArray(16777216, expect);	// 2^24
		expect = new byte[] {(byte)0x01, (byte)0x00, (byte)0x00, (byte)0x01 };
		checkInt2ByteArray(16777217, expect);

		expect = new byte[] {(byte)0x7f, (byte)0xff, (byte)0xff, (byte)0xfe };
		checkInt2ByteArray(2147483646, expect);
		expect = new byte[] {(byte)0x7f, (byte)0xff, (byte)0xff, (byte)0xff };
		checkInt2ByteArray(2147483647, expect);	// 2^31 - 1
	}
	
	public void testPut () {
		Code.put(-1); code.expect(0xff, 1);
		Code.put(0); code.expect(0x00, 1);
		Code.put(1); code.expect(0x01, 1);

		Code.put(127); code.expect(0x7f, 1);
		Code.put(128); code.expect(0x80, 1);
		Code.put(129); code.expect(0x81, 1);

		Code.put(255); code.expect(0xff, 1);
		Code.put(256); code.expect(0x00, 1);
		Code.put(257); code.expect(0x01, 1);
		
		Code.write(code);

		code.setCodeSize(9);
		code.verify();		
	}
	
	public void testPut2 () {
		Code.put2(-1); code.expect(0xffff, 2);
		Code.put2(0); code.expect(0x0000, 2);
		Code.put2(1); code.expect(0x0001, 2);

		Code.put2(127); code.expect(0x007f, 2);
		Code.put2(128); code.expect(0x0080, 2);
		Code.put2(129); code.expect(0x0081, 2);

		Code.put2(255); code.expect(0x00ff, 2);
		Code.put2(256); code.expect(0x0100, 2);
		Code.put2(257); code.expect(0x0101, 2);

		Code.put2(65535); code.expect(0xffff, 2);
		Code.put2(65536); code.expect(0x0000, 2);
		Code.put2(65537); code.expect(0x0001, 2);
		
		Code.write(code);
		
		code.setCodeSize(24);
		code.verify();
	}
	
	public void testPut4 () {
		Code.put4(-1); code.expect(0xffffffff, 4);
		Code.put4(0); code.expect(0x00000000, 4);
		Code.put4(1); code.expect(0x00000001, 4);

		Code.put4(127); code.expect(0x0000007f, 4);
		Code.put4(128); code.expect(0x00000080, 4);
		Code.put4(129); code.expect(0x00000081, 4);

		Code.put4(255); code.expect(0x000000ff, 4);
		Code.put4(256); code.expect(0x00000100, 4);
		Code.put4(257); code.expect(0x00000101, 4);

		Code.put4(65535); code.expect(0x0000ffff, 4);
		Code.put4(65536); code.expect(0x00010000, 4);
		Code.put4(65537); code.expect(0x00010001, 4);
		
		Code.put4(16777215); code.expect(0x00ffffff, 4);
		Code.put4(16777216); code.expect(0x01000000, 4);
		Code.put4(16777217); code.expect(0x01000001, 4);
		
		Code.put4(2147483646); code.expect(0x7ffffffe, 4);
		Code.put4(2147483647); code.expect(0x7fffffff, 4);

		Code.write(code);
		
		code.setCodeSize(68);
		code.verify();
	}
	
	public void testLoad () {
		Item x;

		x = new Item(Item.Local, 4, Tab.intType);
		Code.load(x);
		code.expect(1, 1);	// load
		code.expect(4, 1);	// b
		
		x = new Item(Item.Local, 0, Tab.intType);
		Code.load(x);
		code.expect(2, 1);	// load_0
		
		x = new Item(Item.Local, 3, Tab.intType);
		Code.load(x);
		code.expect(5, 1);	// load_3

		x = new Item(Item.Static, 2, Tab.intType);
		Code.load(x);
		code.expect(11, 1);	// getstatic
		code.expect(2, 2);	// s

		x = new Item(Item.Fld, 1, Tab.intType);
		Code.load(x);
		code.expect(13, 1);	// getfield
		code.expect(1, 2);	// s
		
		x = new Item(Item.Con, 74, Tab.charType);	// 77 = 'J'
		Code.load(x);
		code.expect(22, 1);	// const_
		code.expect(74, 4);	// w
		
		x = new Item(Item.Con, 0, Tab.nullType);
		Code.load(x);
		code.expect(15, 1);	// const_0
				
		x = new Item(Item.Stack, 23, Tab.intType);
		Code.load(x);
		// item is already on stack, no further code necessary
		
		x = new Item(Item.Elem, 23, Tab.intType);
		Code.load(x);
		code.expect(34, 1);	// aload
		
		x = new Item(Item.Elem, 74, Tab.charType);
		Code.load(x);
		code.expect(36, 1);	// baload
		
		Code.write(code);
		
		code.setCodeSize(18);
		code.verify();
	}

	public void testLoadConst () {
		Code.loadConst(0);
		code.expect(15, 1);	// const_0
		
		Code.loadConst(1);
		code.expect(16, 1);	// const_1
		
		Code.loadConst(5);
		code.expect(20, 1);	// const_5
		
		Code.loadConst(-1);
		code.expect(21, 1);	// const_m1
		
		Code.loadConst(23);
		code.expect(22, 1);	// const_
		code.expect(23, 4);	// w
		
		Code.write(code);
		
		code.setCodeSize(9);
		code.verify();
	}
	
	public void testAssign () {
		Item x, y = new Item(Item.Stack, 23, Tab.intType);
		
		x = new Item(Item.Local, 4, Tab.intType);
		Code.assign(x, y);
		code.expect(6, 1);	// store
		code.expect(4, 1);	// b
		
		x = new Item(Item.Local, 0, Tab.intType);
		Code.assign(x, y);
		code.expect(7, 1);	// store_0
		
		x = new Item(Item.Local, 1, Tab.intType);
		Code.assign(x, y);
		code.expect(8, 1);	// store_1
		
		x = new Item(Item.Static, 3, Tab.intType);
		Code.assign(x, y);
		code.expect(12, 1);	// putstatic
		code.expect(3, 2);	// s

		x = new Item(Item.Fld, 0, Tab.intType);
		Code.assign(x, y);
		code.expect(14, 1);	// putfield
		code.expect(0, 2);	// s
		
		x = new Item(Item.Elem, 7, Tab.intType);
		Code.assign(x, y);
		code.expect(35, 1);	// astore
		
		x = new Item(Item.Elem, 8, Tab.charType);
		y.type = Tab.charType;
		Code.assign(x, y);
		code.expect(37, 1);	// bastore
		
		Code.write(code);
		
		code.setCodeSize(12);
		code.verify();
	}
	
	public void testInc () {
		Item x;
		
		x = new Item(Item.Local, 2, Tab.intType);
		Code.inc(x, 4);
		code.expect(31, 1);	// inc
		code.expect(2, 1);	// b1
		code.expect(4, 1);	// b2
		
		x = new Item(Item.Static, 3, Tab.intType);
		Code.inc(x, 5);
		code.expect(11, 1);	// getstatic
		code.expect(3, 2);	// s
		code.expect(20, 1);	// const_5
		code.expect(23, 1);	// add
		code.expect(12, 1);	// putstatic
		code.expect(3, 2);	// s
		
		Code.inc(x, 222);
		code.expect(11, 1);	// getstatic
		code.expect(3, 2);	// s
		code.expect(22, 1);	// const_
		code.expect(222, 4);	// w
		code.expect(23, 1);	// add
		code.expect(12, 1);	// putstatic
		code.expect(3, 2);	// s
		
		x = new Item(Item.Fld, 9, Tab.intType);
		Code.inc(x, 2);
		code.expect(40, 1);	// dup
		code.expect(13, 1);	// getfield
		code.expect(9, 2);	// s
		code.expect(17, 1);	// const_2
		code.expect(23, 1);	// add
		code.expect(14, 1);	// putfield
		code.expect(9, 2);	// s
		
		Code.inc(x, 45);
		code.expect(40, 1);	// dup
		code.expect(13, 1);	// getfield
		code.expect(9, 2);	// s
		code.expect(22, 1);	// const_
		code.expect(45, 4);	// w
		code.expect(23, 1);	// add
		code.expect(14, 1);	// putfield
		code.expect(9, 2);	// s
		
		x = new Item(Item.Elem, 123, Tab.intType);
		Code.inc(x, 0);
		code.expect(41, 1);	// dup2
		code.expect(34, 1);	// aload
		code.expect(15, 1);	// const_0
		code.expect(23, 1);	// add
		code.expect(35, 1);	// astore
		
		Code.inc(x, 45);
		code.expect(41, 1);	// dup2
		code.expect(34, 1);	// aload
		code.expect(22, 1);	// const_
		code.expect(45, 4);	// w
		code.expect(23, 1);	// add
		code.expect(35, 1);	// astore
		
		Code.write(code);
		
		code.setCodeSize(3 + 8 + 12 + 9 + 13 + 5 + 9);
		code.verify();
	}
	
	public static void main(String[] args) {
		junit.textui.TestRunner.run(CodeTest.class);
	}
}
